#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchoutil.core.get import (  # noqa: F401
    CUDA_IF_AVAILABLE,
    DeviceLike,
    DTypeLike,
    GeneratorLike,
    get_device,
    get_dtype,
    get_generator,
)
